@@ -1,5 +1,5 @@
 /*
-Copyright (c) 2018-2019 Genome Research Ltd.
+Copyright (c) 2018-2020 Genome Research Ltd.
 Authors: James Bonfield <jkb@sanger.ac.uk>, Valeriu Ohan <vo2@sanger.ac.uk>
 
 Redistribution and use in source and binary forms, with or without
@@ -136,10 +136,10 @@ static void sam_hrecs_remove_ref_altnames(sam_hrecs_t *hrecs, int expected, cons
  *        -1 on failure
  */
 static int sam_hrecs_update_hashes(sam_hrecs_t *hrecs,
-                                   int type,
+                                   khint32_t type,
                                    sam_hrec_type_t *h_type) {
     /* Add to reference hash? */
-    if ((type>>8) == 'S' && (type&0xff) == 'Q') {
+    if (type == TYPEKEY("SQ")) {
         sam_hrec_tag_t *tag = h_type->tag;
         int nref = hrecs->nref;
         const char *name = NULL;
@@ -177,6 +177,7 @@ static int sam_hrecs_update_hashes(sam_hrecs_t *hrecs,
         k = kh_get(m_s2i, hrecs->ref_hash, name);
         if (k < kh_end(hrecs->ref_hash)) {
             nref = kh_val(hrecs->ref_hash, k);
+            int ref_changed_flag = 0;
 
             // Check for hash entry added by sam_hrecs_refs_from_targets_array()
             if (hrecs->ref[nref].ty == NULL) {
@@ -189,23 +190,30 @@ static int sam_hrecs_update_hashes(sam_hrecs_t *hrecs,
                              hrecs->ref[nref].len);
                     if (sam_hrecs_update(hrecs, h_type, "LN", tmp, NULL) < 0)
                         return -1;
+                    ref_changed_flag = 1;
                 }
                 if (sam_hrecs_add_ref_altnames(hrecs, nref, altnames) < 0)
                     return -1;
 
-                if (hrecs->refs_changed < 0 || hrecs->refs_changed > nref)
+                if (ref_changed_flag && (hrecs->refs_changed < 0 || hrecs->refs_changed > nref))
                     hrecs->refs_changed = nref;
                 return 0;
             }
 
             // Check to see if an existing entry is being updated
             if (hrecs->ref[nref].ty == h_type) {
-                hrecs->ref[nref].len = len;
-                hrecs->ref[nref].name = name;
+                if (hrecs->ref[nref].len != len) {
+                    hrecs->ref[nref].len = len;
+                    ref_changed_flag = 1;
+                }
+                if (!hrecs->ref[nref].name || strcmp(hrecs->ref[nref].name, name)) {
+                    hrecs->ref[nref].name = name;
+                    ref_changed_flag = 1;
+                }
                 if (sam_hrecs_add_ref_altnames(hrecs, nref, altnames) < 0)
                     return -1;
 
-                if (hrecs->refs_changed < 0 || hrecs->refs_changed > nref)
+                if (ref_changed_flag && (hrecs->refs_changed < 0 || hrecs->refs_changed > nref))
                     hrecs->refs_changed = nref;
                 return 0;
             }
@@ -251,7 +259,7 @@ static int sam_hrecs_update_hashes(sam_hrecs_t *hrecs,
     }
 
     /* Add to read-group hash? */
-    if ((type>>8) == 'R' && (type&0xff) == 'G') {
+    if (type == TYPEKEY("RG")) {
         sam_hrec_tag_t *tag = sam_hrecs_find_key(h_type, "ID", NULL);
         int nrg = hrecs->nrg, r;
         khint_t k;
@@ -299,7 +307,7 @@ static int sam_hrecs_update_hashes(sam_hrecs_t *hrecs,
     }
 
     /* Add to program hash? */
-    if ((type>>8) == 'P' && (type&0xff) == 'G') {
+    if (type == TYPEKEY("PG")) {
         sam_hrec_tag_t *tag;
         sam_hrec_pg_t *new_pg;
         int npg = hrecs->npg;
@@ -384,7 +392,7 @@ static int sam_hrecs_update_hashes(sam_hrecs_t *hrecs,
     return 0;
 }
 
-static int sam_hrecs_remove_hash_entry(sam_hrecs_t *hrecs, int type, sam_hrec_type_t *h_type) {
+static int sam_hrecs_remove_hash_entry(sam_hrecs_t *hrecs, khint32_t type, sam_hrec_type_t *h_type) {
     if (!hrecs || !h_type)
         return -1;
 
@@ -393,7 +401,7 @@ static int sam_hrecs_remove_hash_entry(sam_hrecs_t *hrecs, int type, sam_hrec_ty
     khint_t k;
 
     /* Remove name and any alternative names from reference hash */
-    if ((type>>8) == 'S' && (type&0xff) == 'Q') {
+    if (type == TYPEKEY("SQ")) {
         const char *altnames = NULL;
 
         tag = h_type->tag;
@@ -433,7 +441,7 @@ static int sam_hrecs_remove_hash_entry(sam_hrecs_t *hrecs, int type, sam_hrec_ty
     }
 
     /* Remove from read-group hash */
-    if ((type>>8) == 'R' && (type&0xff) == 'G') {
+    if (type == TYPEKEY("RG")) {
         tag = h_type->tag;
 
         while (tag) {
@@ -474,7 +482,7 @@ static int sam_hrecs_remove_hash_entry(sam_hrecs_t *hrecs, int type, sam_hrec_ty
 static void sam_hrecs_global_list_add(sam_hrecs_t *hrecs,
                                       sam_hrec_type_t *h_type,
                                       sam_hrec_type_t *after) {
-    const khint32_t hd_type = 'H' << 8 | 'D';
+    const khint32_t hd_type = TYPEKEY("HD");
     int update_first_line = 0;
 
     // First line seen
@@ -528,7 +536,7 @@ static int sam_hrecs_vadd(sam_hrecs_t *hrecs, const char *type, va_list ap, ...)
     sam_hrec_type_t *h_type;
     sam_hrec_tag_t *h_tag, *last=NULL;
     int new;
-    khint32_t type_i = (type[0]<<8) | type[1], k;
+    khint32_t type_i = TYPEKEY(type), k;
 
     if (!strncmp(type, "HD", 2) && (h_type = sam_hrecs_find_type_id(hrecs, "HD", NULL, NULL)))
         return sam_hrecs_vupdate(hrecs, h_type, ap);
@@ -640,8 +648,7 @@ static int sam_hrecs_vadd(sam_hrecs_t *hrecs, const char *type, va_list ap, ...)
         last = h_tag;
     }
 
-    int itype = (type[0]<<8) | type[1];
-    if (-1 == sam_hrecs_update_hashes(hrecs, itype, h_type))
+    if (-1 == sam_hrecs_update_hashes(hrecs, TYPEKEY(type), h_type))
         return -1;
 
     if (!strncmp(type, "PG", 2))
@@ -679,7 +686,7 @@ static int sam_hrecs_remove_line(sam_hrecs_t *hrecs, const char *type_name, sam_
     if (!hrecs || !type_name || !type_found)
         return -1;
 
-    int itype = (type_name[0]<<8) | type_name[1];
+    khint32_t itype = TYPEKEY(type_name);
     khint_t k = kh_get(sam_hrecs_t, hrecs->h, itype);
     if (k == kh_end(hrecs->h))
         return -1;
@@ -778,12 +785,12 @@ static int sam_hrecs_parse_lines(sam_hrecs_t *hrecs, const char *hdr, size_t len
             return -1;
         }
 
-        type = (((uint8_t) hdr[i+1])<<8) | (uint8_t) hdr[i+2];
         if (!isalpha_c(hdr[i+1]) || !isalpha_c(hdr[i+2])) {
             sam_hrecs_error("Header line does not have a two character key",
                           &hdr[l_start], len - l_start, lno);
             return -1;
         }
+        type = TYPEKEY(&hdr[i+1]);
 
         i += 3;
         if (i == len || hdr[i] == '\n')
@@ -819,7 +826,7 @@ static int sam_hrecs_parse_lines(sam_hrecs_t *hrecs, const char *hdr, size_t len
 
         // Parse the tags on this line
         last = NULL;
-        if ((type>>8) == 'C' && (type&0xff) == 'O') {
+        if (type == TYPEKEY("CO")) {
             size_t j;
 
             if (i == len || hdr[i] != '\t') {
@@ -1237,8 +1244,10 @@ int sam_hdr_rebuild(sam_hdr_t *bh) {
     if (!hrecs->dirty)
         return 0;
 
-    if (hrecs->pgs_changed)
-        sam_hdr_link_pg(bh);
+    if (hrecs->pgs_changed && sam_hdr_link_pg(bh) < 0) {
+        hts_log_error("Linking @PG lines has failed");
+        return -1;
+    }
 
     kstring_t ks = KS_INITIALIZE;
     if (sam_hrecs_rebuild_text(hrecs, &ks) != 0) {
@@ -1616,8 +1625,7 @@ int sam_hdr_remove_except(sam_hdr_t *bh, const char *type, const char *ID_key, c
 
     sam_hrec_type_t *type_found = sam_hrecs_find_type_id(hrecs, type, ID_key, ID_value);
     if (!type_found) { // remove all line of this type
-        int itype = (type[0]<<8)|(type[1]);
-        khint_t k = kh_get(sam_hrecs_t, hrecs->h, itype);
+        khint_t k = kh_get(sam_hrecs_t, hrecs->h, TYPEKEY(type));
         if (k == kh_end(hrecs->h))
             return 0;
         type_found =  kh_val(hrecs->h, k);
@@ -1642,9 +1650,9 @@ int sam_hdr_remove_except(sam_hdr_t *bh, const char *type, const char *ID_key, c
     return 0;
 }
 
-int sam_hdr_remove_lines(sam_hdr_t *bh, const char *type, const char *id, void *h) {
+int sam_hdr_remove_lines(sam_hdr_t *bh, const char *type, const char *id, void *vrh) {
     sam_hrecs_t *hrecs;
-    rmhash_t *rh = (rmhash_t *)h;
+    rmhash_t *rh = (rmhash_t *)vrh;
 
     if (!bh || !type)
         return -1;
@@ -1659,8 +1667,7 @@ int sam_hdr_remove_lines(sam_hdr_t *bh, const char *type, const char *id, void *
         hrecs = bh->hrecs;
     }
 
-    int itype = (type[0]<<8)|(type[1]);
-    khint_t k = kh_get(sam_hrecs_t, hrecs->h, itype);
+    khint_t k = kh_get(sam_hrecs_t, hrecs->h, TYPEKEY(type));
     if (k == kh_end(hrecs->h)) // nothing to remove from
         return 0;
 
@@ -2020,7 +2027,7 @@ hts_pos_t sam_hdr_tid2len(const sam_hdr_t *h, int tid) {
 
 /*
  * Fixes any PP links in @PG headers.
- * If the entries are in order then this doesn't need doing, but incase
+ * If the entries are in order then this doesn't need doing, but in case
  * our header is out of order this goes through the hrecs->pg[] array
  * setting the prev_id field.
  *
@@ -2044,7 +2051,7 @@ static int sam_hdr_link_pg(sam_hdr_t *bh) {
         hrecs = bh->hrecs;
     }
 
-    if (!hrecs->pgs_changed)
+    if (!hrecs->pgs_changed || !hrecs->npg)
         return 0;
 
     hrecs->npg_end_alloc = hrecs->npg;
@@ -2052,6 +2059,9 @@ static int sam_hdr_link_pg(sam_hdr_t *bh) {
     if (!new_pg_end)
         return -1;
     hrecs->pg_end = new_pg_end;
+    int *chain_size = calloc(hrecs->npg, sizeof(int));
+    if (!chain_size)
+        return -1;
 
     for (i = 0; i < hrecs->npg; i++)
         hrecs->pg_end[i] = i;
@@ -2073,24 +2083,33 @@ static int sam_hdr_link_pg(sam_hdr_t *bh) {
         k = kh_get(m_s2i, hrecs->pg_hash, tag->str+3);
 
         if (k == kh_end(hrecs->pg_hash)) {
-            ret = -1;
+            hts_log_warning("PG line with PN:%s has a PP link to missing program '%s'",
+                    hrecs->pg[i].name, tag->str+3);
             continue;
         }
 
         hrecs->pg[i].prev_id = hrecs->pg[kh_val(hrecs->pg_hash, k)].id;
         hrecs->pg_end[kh_val(hrecs->pg_hash, k)] = -1;
+        chain_size[i] = chain_size[kh_val(hrecs->pg_hash, k)]+1;
     }
 
     for (i = j = 0; i < hrecs->npg; i++) {
-        if (hrecs->pg_end[i] != -1)
+        if (hrecs->pg_end[i] != -1 && chain_size[i] > 0)
             hrecs->pg_end[j++] = hrecs->pg_end[i];
     }
+    /* Only leafs? Choose the last one! */
+    if (!j && hrecs->npg_end > 0) {
+        hrecs->pg_end[0] = hrecs->pg_end[hrecs->npg_end-1];
+        j = 1;
+    }
+
     hrecs->npg_end = j;
     hrecs->pgs_changed = 0;
 
     /* mark as dirty or empty for rebuild */
     hrecs->dirty = 1;
     redact_header_text(bh);
+    free(chain_size);
 
     return ret;
 }
@@ -2165,6 +2184,12 @@ int sam_hdr_add_pg(sam_hdr_t *bh, const char *name, ...) {
         hrecs = bh->hrecs;
     }
 
+    bh->hrecs->pgs_changed = 1;
+    if (sam_hdr_link_pg(bh) < 0) {
+        hts_log_error("Error linking @PG lines");
+        return -1;
+    }
+
     va_list args;
     // Check for ID / PN / PP tags in varargs list
     va_start(args, name);
@@ -2403,7 +2428,6 @@ sam_hrec_type_t *sam_hrecs_find_type_id(sam_hrecs_t *hrecs, const char *type,
     if (!hrecs || !type)
         return NULL;
     sam_hrec_type_t *t1, *t2;
-    int itype = (type[0]<<8)|(type[1]);
     khint_t k;
 
     /* Special case for types we have prebuilt hashes on */
@@ -2436,7 +2460,7 @@ sam_hrec_type_t *sam_hrecs_find_type_id(sam_hrecs_t *hrecs, const char *type,
         }
     }
 
-    k = kh_get(sam_hrecs_t, hrecs->h, itype);
+    k = kh_get(sam_hrecs_t, hrecs->h, TYPEKEY(type));
     if (k == kh_end(hrecs->h))
         return NULL;
 
